const config_struct = require("../data-struct/config-data-struct");
const logger = require("pomelo-logger").getLogger("redis", __filename);

let redisPool = undefined;

/**
 * 空方法
 */
function emptyFunc(error, result) { };

/**
 * 操作redis的封装对象
 * 
 * @class RedisOperateClass
 */
class RedisOperateClass {

    /**
     * 构造函数
     * @param {config_struct} redisConfig redis配置
     * @param {Function} callback 回调函数
     * @memberof RedisOperateClass
     */
    constructor(redisConfig, callback) {

        try {

            const RDS_PORT = redisConfig.port;
            const RDS_HOST = redisConfig.host;
            const RDS_OPTS = redisConfig.opts;
            const MAX_COUNT = redisConfig.maxcount || 5;
            const PERFORM_CHECKS = redisConfig.perform_checks;
            const sharding_count = redisConfig.sharding_count || 16;

            RDS_OPTS.retry_strategy = function (options) {

                if (options.error.code === 'ECONNREFUSED') {
                    // End reconnecting on a specific error and flush all commands with a individual error 
                    logger.error('Redis连接被拒绝');
                }

                if (options.error.code === 'ETIMEDOUT') {
                    logger.error(options.error.message);
                }

                if (options.timesConnected > 10) {
                    logger.error('Redis重试连接超过十次');
                }
                // reconnect after 
                return Math.max(options.attempt * 100, 3000);
            };

            let initedCount = 0;

            RDS_OPTS.createSuccess = (database) => {
                initedCount++;
                logger.debug("redis client database %d 初始化完毕", database);
                if (initedCount === sharding_count) {
                    callback && (callback());
                }
            };

            RDS_OPTS.createError = (database, error) => {
                logger.error("redis client database %d 初始化失败:%s", database, error);
            };

            const pool_creator = require('redis-connection-pool');

            redisPool = new Array(sharding_count);

            for (var index = 0; index < sharding_count; index++) {

                const sharding_uid = `redis_sharding_${index}`;

                redisPool[index] = pool_creator(
                    sharding_uid,
                    {
                        host: RDS_HOST,                 // 主机地址 
                        port: RDS_PORT,                 // 端口 
                        max_clients: MAX_COUNT,         // 最大连接数 
                        perform_checks: PERFORM_CHECKS, // 检查是否启用推送/取出功能
                        database: index,                // 分片
                        options: RDS_OPTS,              // 创建node-redis客户端的选项，可选参数
                    }
                );
            }

        } catch (error) {
            throw error;
        }
    }

    /**
     * 设置key的失效时间
     * 
     * @param {string} key          key
     * @param {number} timeout      超时时间
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 
     * @memberof RedisOperateClass
     */
    expire(key, timeout, index, callback) {
        redisPool[index].expire(key, timeout, callback ? callback : emptyFunc);
    }


    /**
     * 获取key剩余时间
     *
     * @param {string} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数
     * @memberof RedisOperateClass
     */
    ttl(key, index, callback) {
        redisPool[index].ttl(key, callback ? callback : emptyFunc);
    }

    /**
     * 获取哈希表长度
     *
     * @param {String} key          哈希表key
     * @param {number} index        片区下标
     * @param {Function} callback   回调函数 参数 1.错误信息 2.哈希表长度
     * @memberof RedisOperateClass
     */
    hlen(key, index, callback) {
        redisPool[index].hlength(key, callback ? callback : emptyFunc);
    }

    /**
     * 发送redis命令
     *
     * @param {String} commond         命令
     * @param {String[]} args       参数数组
     * @param {number} index        片区下标
     * @param {Function} callback   回调函数 参数 1.错误信息 2.哈希表长度
     * @memberof RedisOperateClass
     */
    sendCommond(commond, args, index, callback) {
        redisPool[index].send_command(commond, args, callback ? callback : emptyFunc);
    }

    /**
     * 是否存在key,0不存在，1存在
     *
     * @param {String} key          key
     * @param {Number} index        片区下标
     * @param {Function} callback   回调函数 参数 1.错误信息 2.哈希表长度
     * @memberof RedisOperateClass
     */
    exists(key, index, callback) {
        this.sendCommond("exists", [key], index, callback ? callback : emptyFunc);
    }

    /**
     * 字符串set操作
     * 
     * @param {String} key          key          
     * @param {String} value        值             
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 1.错误信息 2.设置的个数
     */
    stringSet(key, value, index, callback) {
        redisPool[index].set(key, value, callback ? callback : emptyFunc);
    }

    /**
     * 多条字符串的set操作
     * 
     * @param {[String]} keys        key数组
     * @param {[String]} values      值数组
     * @param {number} index         分片索引
     * @param {Function} callback    回调函数 1.错误信息 2.设置的个数 
     */
    stringMSet(keys, values, index, callback) {
        redisPool[index].mset(keys, values, callback ? callback : emptyFunc);
    }

    /**
     * 字符串Get操作
     * 
     * @param {String} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 1.错误信息 2.获取到的字符串值
     */
    stringGet(key, index, callback) {
        redisPool[index].get(key, callback ? callback : emptyFunc);
    }


    /**
     * 多条字符串Get操作
     * 
     * @param {[String]} keys        key数组
     * @param {number} index         分片索引
     * @param {Function} callback    回调函数 1.错误信息 2.获取到的字符串数组
     */
    stringMGet(keys, index, callback) {
        redisPool[index].mget(keys, callback ? callback : emptyFunc);
    }

    /**
     * 根据key删除字符串
     * 
     * @param {String} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数
     */
    stringDel(key, index, callback) {
        redisPool[index].del(key, callback ? callback : emptyFunc);
    }


    /**
     * 插入或更新数据到哈希表指定字段
     * 
     * 如果 field 是哈希表中的一个新建域，并且值设置成功，返回 1 。
     * 如果哈希表中域 field 已经存在且旧值已被新值覆盖，返回 0 。
     * 
     * @param {String} key          哈表的key
     * @param {String} field        哈希表中字段的key
     * @param {String} value        值
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数     
     */
    hSet(key, field, value, index, callback) {
        redisPool[index].hset(key, field, value, callback ? callback : emptyFunc);
    }

    /**
     * 插入多条数据到哈希表,如果命令执行成功，返回 "OK"。
     * 
     * @param {String} key          哈表的key
     * @param {[String]} fields     哈希表中字段的key数组
     * @param {[String]} values     值数组
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数    
     */
    hMset(key, fields, values, index, callback) {
        redisPool[index].hmset(key, fields, values, callback ? callback : emptyFunc);
    }

    /**
     * 递增指定哈希表字段值
     *
     * @param {String} key          哈希表的key
     * @param {String} field        哈希表中字段的key
     * @param {number} value        值
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.修改后的结果
     */
    hIncrby(key, field, value, index, callback) {
        redisPool[index].hincrby(key, field, value, callback ? callback : emptyFunc);
    }

    /**
     * 获取哈希表中的单条数据,如果没有该数据，则返回null
     * 
     * @param {String} key          哈希表的key
     * @param {String} field        哈希表中字段的key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.数据 
     */
    hGet(key, field, index, callback) {
        redisPool[index].hget(key, field, callback ? callback : emptyFunc);
    }

    /**
     * 获取哈希表中的多条数据
     * 一个包含多个给定域的关联值的表，表值的排列顺序和给定域参数的请求顺序一样。
     * 不存在的域返回null值。
     *
     * @param {String} key          哈希表的key
     * @param {[String]} fields     哈希表中字段的key的数组
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.数据数组
     */
    hMget(key, fields, index, callback) {
        redisPool[index].hmget(key, fields, callback ? callback : emptyFunc);
    }

    /**
     * 获取指定哈希表所有数据
     * 
     * @param {String} key          哈希表的key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.数据数组
     */
    hGetAll(key, index, callback) {
        redisPool[index].hgetall(key, callback ? callback : emptyFunc);
    }

    /**
     * 删除哈希表 key 中的一个或多个指定域，不存在的域将被忽略。
     * 
     * 返回被成功移除的域的数量，不包括被忽略的域。
     * 
     * @param {String} key          哈希表的key
     * @param {[String]} fields     哈希表中字段的key的数组
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.删除的数据数量
     */
    hDel(key, fields, index, callback) {
        redisPool[index].hdel(key, fields, callback ? callback : emptyFunc);
    }

    /**
     * 获取哈希表中的所有字段名
     * 
     * @param {String} key          哈希表的key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.结果
     */
    hKeys(key, index, callback) {
        redisPool[index].hkeys(key, callback ? callback : emptyFunc);
    }

    /**
     * 查看哈希表的指定字段是否存在
     * 
     * @param {String} key          哈希表的key
     * @param {String} field        哈希表keyId
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.结果
     */
    hExists(key, field, index, callback) {
        redisPool[index].hexists(key, field, callback ? callback : emptyFunc);
    }

    /**
     * 向存于 key 的列表的尾部插入所有指定的值
     *
     * @param {String} key          key
     * @param {String} value        值
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.操作后的列表长度
     */
    rpush(key, value, index, callback) {
        redisPool[index].rpush(key, value, callback ? callback : emptyFunc);
    }

    /**
     * 将指定的值插入到存于 key 的列表的头部
     * 
     * @param {String} key          key
     * @param {String} value        值
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.操作后的列表长度
     */
    lpush(key, value, index, callback) {
        redisPool[index].lpush(key, value, callback ? callback : emptyFunc);
    }

    /**
     * 移除并返回存于 key 的 list 的最后一个元素
     * 
     * @param {String} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.最后一个元素的值
     */
    rpop(key, index, callback) {
        redisPool[index].brpop(key, callback ? callback : emptyFunc);
    }

    /**
     * 移除并且返回 key 对应的 list 的第一个元素
     * 
     * @param {String} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.返回第一个元素的值
     */
    lpop(key, index, callback) {
        redisPool[index].blpop(key, callback ? callback : emptyFunc);
    }

    /**
     * 返回存储在 key 的列表里指定范围内的元素。 
     * start 和 end 偏移量都是基于0的下标，即
     * list的第一个元素下标是0（list的表头），
     * 第二个元素下标是1，以此类推。偏移量也可以
     * 是负数，表示偏移量是从list尾部开始计数。 
     * 例如， -1 表示列表的最后一个元素，
     * -2 是倒数第二个，以此类推。
     *
     * @param {String} key          key
     * @param {number} start        起始位置
     * @param {number} end          结束位置
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.指定范围里的列表元素
     */
    lrange(key, start, end, index, callback) {
        redisPool[index].blrange(key, start, end, callback ? callback : emptyFunc);
    }

    /**
     * 返回存储在 key 里的list的长度
     * 
     * @param {String} key          key
     * @param {number} index        分片索引
     * @param {Function} callback   回调函数 参数 1.错误信息 2.key对应的list的长度
     */
    llen(key, index, callback) {
        redisPool[index].bllen(key, callback ? callback : emptyFunc);
    }
}

module.exports = RedisOperateClass;